import collections

import tensorflow as tf
import numpy as np
import math as m

from Neural_Networks import vae_utils

# based on implementation here:
# https://github.com/tensorflow/models/blob/master/autoencoder/autoencoder_models/VariationalAutoencoder.py

SMALL_CONSTANT = 1e-6

class VariationalAutoencoder(object):

    def __init__(self, name, n_input, n_hidden, n_weights, middle="gaussian"):
        
        self.n_input = n_input
        self.n_hidden = n_hidden
        self.n_weights = n_weights
        self.name = name
        self.middle = middle
        self.bias_start = 0.0

        network_weights = self._create_weights()
        self.weights = network_weights

        self.nonlinearity = tf.nn.relu
        self.nonlinearity2 = tf.nn.leaky_relu


    def _calc_z_mean_and_sigma(self,x):
        with tf.name_scope("VICI_encoder"):

            # literally multiplies input layer by weights matrix and adds bias vector
            layer_1_pre = tf.add(tf.matmul(x, self.weights['VICI_encoder']['W_layer_1']), self.weights['VICI_encoder']['b_layer_1'])
            previous_layer_out = self.nonlinearity(layer_1_pre) # passes the output through the activation

            # literally multiplies output of previous layer by weights matrix and adds bias vector
            layer_2_pre = tf.add(tf.matmul(previous_layer_out, self.weights['VICI_encoder']['W_layer_2']), self.weights['VICI_encoder']['b_layer_2'])
            previous_layer_out = self.nonlinearity(layer_2_pre)

            # literally multiplies input layer by weights matrix and adds bias vector - no activation (final output)
            z_mean = tf.add(tf.matmul(previous_layer_out, self.weights['VICI_encoder']['W_layer_out_mu']), self.weights['VICI_encoder']['b_layer_out_mu'])
            tf.summary.histogram("z_mean", z_mean)

            # literally multiplies input layer by weights matrix and adds bias vector - no activation (final output)
            z_log_sigma_sq = tf.add(tf.matmul(previous_layer_out, self.weights['VICI_encoder']['W_layer_out_log_sigma']), self.weights['VICI_encoder']['b_layer_out_log_sigma'])
            tf.summary.histogram("z_log_sigma_sq", z_log_sigma_sq)
            
            return z_mean, z_log_sigma_sq

    def _sample_from_gaussian_dist(self, num_rows, num_cols, mean, log_sigma_sq):
        with tf.name_scope("sample_in_z_space"):
            eps = tf.random_normal([num_rows, num_cols], 0, 1., dtype=tf.float32)
            sample = tf.add(mean, tf.multiply(tf.sqrt(tf.exp(log_sigma_sq)), eps))
        return sample
    
#    def _sample_from_gaussian_conditional_dist(self, num_rows, num_cols, mean, x_multi, log_sigma_sq):
#        with tf.name_scope("sample_from_x_distribution"):
#            dp = tf.zeros([tf.shape(mean)[0],1],tf.float32)
##            sample = tf.zeros([tf.shape(mean)[0],tf.shape(mean)[1]],tf.float32)
#            sample = tf.zeros([tf.shape(mean)[0],0],dtype=tf.float32)
#            eps = tf.random_normal([num_rows, num_cols], 0, 1., dtype=tf.float32)
#            for i in range(num_cols):
#                mean_i = tf.expand_dims(mean[:,i],axis=1)
#                log_sigma_sq_i = tf.expand_dims(log_sigma_sq[:,i],axis=1)
#                eps_i = tf.expand_dims(eps[:,i],axis=1)
#                x_multi_i = tf.expand_dims(x_multi[:,i],axis=1)
##                eps = tf.random_normal([num_rows,1], 0, 1., dtype=tf.float32)
#                dp = tf.add(mean_i+tf.multiply(dp,x_multi_i), tf.multiply(tf.sqrt(tf.exp(log_sigma_sq_i)), eps_i))
##                sample[:,i] = dp
#                sample = tf.concat([sample,dp],1)
#                
#        return sample

    def _create_weights(self):
        all_weights = collections.OrderedDict()
        with tf.variable_scope("VICI_ENC"):
            
            # Encoder
            all_weights['VICI_encoder'] = collections.OrderedDict()
            hidden_number_encoder = self.n_weights

            # initiate a matrix of weights for the first layer - size of input X n_weights (bias vector length n_weights) 
            all_weights['VICI_encoder']['W_layer_1'] = tf.Variable(vae_utils.xavier_init(self.n_input, hidden_number_encoder), dtype=tf.float32)
            all_weights['VICI_encoder']['b_layer_1'] = tf.Variable(tf.zeros([hidden_number_encoder], dtype=tf.float32) * self.bias_start) # bad code multiply zero by bias_start
            tf.summary.histogram("W_layer_1", all_weights['VICI_encoder']['W_layer_1'])
    
            # initiate a matrix of weights for the second layer - size of n_weights X n_weights (bias vector length n_weights)
            all_weights['VICI_encoder']['W_layer_2'] = tf.Variable(vae_utils.xavier_init(hidden_number_encoder, hidden_number_encoder), dtype=tf.float32)
            all_weights['VICI_encoder']['b_layer_2'] = tf.Variable(tf.zeros([hidden_number_encoder], dtype=tf.float32) * self.bias_start)
            tf.summary.histogram("W_layer_2", all_weights['VICI_encoder']['W_layer_2'])
                    
            # initiate a matrix of weights for the final mu layer - size of n_weights X output (bias vector length of output)
            all_weights['VICI_encoder']['W_layer_out_mu'] = tf.Variable(vae_utils.xavier_init(hidden_number_encoder, self.n_hidden),dtype=tf.float32)
            all_weights['VICI_encoder']['b_layer_out_mu'] = tf.Variable(tf.zeros([self.n_hidden], dtype=tf.float32) * self.bias_start, dtype=tf.float32)
            tf.summary.histogram("W_layer_out_mu", all_weights['VICI_encoder']['W_layer_out_mu'])
    
            # initiate a matrix of weights for the final log_sig_sq layer - size of n_weights X output (bias vector length of output)
            all_weights['VICI_encoder']['W_layer_out_log_sigma'] = tf.Variable(vae_utils.xavier_init(hidden_number_encoder, self.n_hidden), dtype=tf.float32)
            all_weights['VICI_encoder']['b_layer_out_log_sigma'] = tf.Variable(tf.zeros([self.n_hidden], dtype=tf.float32) * self.bias_start, dtype=tf.float32)
            tf.summary.histogram("W_layer_out_log_sigma", all_weights['VICI_encoder']['W_layer_out_log_sigma'])
    
            all_weights['prior_param'] = collections.OrderedDict()
        
        return all_weights
